/*
    Name:       DDNSClient.ino
    Created:    2018/8/29 18:34:54 by sy2k.com
    Modified:   2018/8/30 add DDNS functions
*/

// Define User Types below here or use a .h file
#include <OLED.h>
#include <TimeLib.h>
#include <WiFiUdp.h>
#include <ESP8266WiFi.h>
#include <ESP8266HTTPClient.h>
#include <WiFiClientSecure.h>

/* OLED value defines */
#define OLED_PIN_RST 16
#define OLED_PIN_SDA 4
#define OLED_PIN_SCL 5
OLED display(OLED_PIN_SDA, OLED_PIN_SCL);
char* charBufp;

/* Wifi value defines */
char ssid[] = "--------";        // your network SSID (name)
char pass[] = "--------";        // your network password
IPAddress ipAddr_local(0, 0, 0, 0);
IPAddress ipAddr_public(0, 0, 0, 0);
IPAddress ipAddr_public_old(0, 0, 0, 0);

/* NTP Settings */
WiFiUDP ntpUDP;                        // A UDP instance to let us send and receive packets over UDP
const char* ntpServerName = "ntp.sjtu.edu.cn";
IPAddress ntpTimeServerIP;            // time.nist.gov NTP server address
#define NTP_LOCAL_PORT        2390    // local port to listen for UDP packets
#define NTP_PACKET_SIZE        48        // NTP time stamp is in the first 48 bytes of the message
byte ntpPacketBuffer[NTP_PACKET_SIZE];    //buffer to hold incoming and outgoing packets
unsigned long utpTimestamp = 0;
unsigned long unixTimestamp = 0;
#define NTP_UNIX_TS_DELTA    2208988800UL    // Unix time starts on Jan 1 1970. In seconds, that's 2208988800
#define TIME_ZONE 8        // setting time zone +8 for Beijing
#define TIME_ZONEADJ_SEC (TIME_ZONE*3600)

/* HTTP for public IP Settings*/
HTTPClient cipHttp;
// const char* cipServerName = "http://members.3322.org/dyndns/getip";
const char* cipServerName = "http://dynamicdns.park-your-domain.com/getip";
String cipHttpStr;

/* DDNS Settings */
#define HTTPS_PORT 443
const char* ddnsServer = "dynamicdns.park-your-domain.com";
const char* ddnsServerFingerprint = "b1 7a b0 ab 23 ac b6 62 7f 36 f2 b4 20 27 12 fe ad 28 d3 b1";
const char* ddnsHost = "--------";        // your namecheap Host
const char* ddnsDomain = "--------";        // your namecheap Domain
const char* ddnsPassword = "--------";        // your namecheap Dynamic DNS Password
WiFiClientSecure ddnsClient;

/* Function execution cycle settings */
unsigned long funcTCnext_timeDisp = 0;
unsigned long funcTCcycle_timeDisp = 1000;
unsigned long funcTCnext_ipSet = 0;
unsigned long funcTCcycle_ipSet = 60000;
// millis() will overflow (go back to zero), after approximately 50 days.
// In order to avoid overflow errors, add a mark when millis() > half ulong
bool millisHFM = false;        // HFM means half full mark
unsigned long funcTCnext_millisHFM = 0;
unsigned long funcTCcycle_millisHFM = 1000;
const unsigned long millisHF_UpThreshold = 0xC000000000000000;
const unsigned long millisHF_HalfThreshold = 0x8000000000000000;
const unsigned long millisHF_DownThreshold = 0x4000000000000000;


// The setup() function runs once each time the micro-controller starts
void setup()
{
    /* System Init */
    Serial.begin(115200);
    Serial.println();
    Serial.println();

    /* OLED Init */
    Serial.print("OLED Init... ");
    pinMode(OLED_PIN_RST, OUTPUT);
    digitalWrite(OLED_PIN_RST, LOW);    // turn D2 low to reset OLED
    delay(50);
    digitalWrite(OLED_PIN_RST, HIGH);    // while OLED is running, must set D2 in high

    display.begin();
    display.clear();
    Serial.println("OK!");

    /* WIFI Init & Connect */
    Serial.print("Connecting to ");
    Serial.println(ssid);
    WiFi.begin(ssid, pass);

    while (WiFi.status() != WL_CONNECTED) {
        delay(500);
        Serial.print(".");
    }
    Serial.println("");
    Serial.println("WiFi connected");

    Serial.print("MAC: ");
    Serial.println(WiFi.macAddress());

    /* NTP UDP starter */
    Serial.println("Starting NTP UDP client");
    ntpUDP.begin(NTP_LOCAL_PORT);
    Serial.print("Local port: ");
    Serial.println(ntpUDP.localPort());
}


// Add the main program code into the continuous loop() function
void loop()
{
    if ((!millisHFM && millis() > funcTCnext_ipSet) || (millisHFM && (millis() + millisHF_HalfThreshold) > (funcTCnext_ipSet - millisHF_HalfThreshold)))
    {
        funcTCnext_ipSet = millis() + funcTCcycle_ipSet;

        /* New loop circle */
        Serial.println("\n/* --- New ipSet circle --- */");
        Serial.print("Running time in milisec = ");
        Serial.println(millis());

        /* NTP Time Update */
        WiFi.hostByName(ntpServerName, ntpTimeServerIP);
        ntpSendPacket(ntpTimeServerIP);        // send an NTP packet to a time server
        delay(500);

        int ntpRepacketLen = ntpUDP.parsePacket();
        if (!ntpRepacketLen) Serial.println("no packet yet");
        else
        {
            Serial.print("packet received, length = ");
            Serial.println(ntpRepacketLen);
            // We've received a packet, read the data from it
            ntpUDP.read(ntpPacketBuffer, NTP_PACKET_SIZE); // read the packet into the buffer

            //the timestamp starts at byte 40 of the received packet and is four bytes,
            // combine the four bytes (two words) into a long integer, this is NTP time (seconds since Jan 1 1900):
            utpTimestamp = ntpPacketBuffer[40] << 24 | ntpPacketBuffer[41] << 16 | ntpPacketBuffer[42] << 8 | ntpPacketBuffer[43];
            unixTimestamp = utpTimestamp - NTP_UNIX_TS_DELTA;
            Serial.printf("Seconds since Jan 1 1900 = %u\n", utpTimestamp);
            Serial.printf("Unix time = %d\n", unixTimestamp);
            setTime(unixTimestamp);
            adjustTime(TIME_ZONEADJ_SEC);

            charBufp = time2char();
            Serial.printf("Time = %s\n", charBufp);
            display.print(charBufp, 0, 0);

            Serial.print("NTP Remote IP address: ");
            Serial.println(ntpUDP.remoteIP());
            Serial.print("NTP Destination IP address: ");
            Serial.println(ntpUDP.destinationIP());
        }

        /* Update local IP */
        ipAddr_local = WiFi.localIP();
        charBufp = ip2char(ipAddr_local);
        Serial.print("Local IP: ");
        Serial.println(charBufp);
        display.print(charBufp, 1, 0);

        /* HTTP for public IP */
        Serial.println("HTTP for public IP begin...");
        cipHttp.begin(cipServerName);
        int httpCode = cipHttp.GET();

        if (httpCode == HTTP_CODE_OK)
        {
            cipHttpStr = cipHttp.getString();
            Serial.println("[HTTP] GET... OK");
            Serial.printf("HTTP String length: %d\n", cipHttpStr.length());
            Serial.println(cipHttpStr);
            // http return with \n end, it cannot be recognized by IPAddress.fromString
            // IPAddress.fromString will change the right part of addr even return false
            cipHttpStr.trim();
            IPAddress ipAddrTemp;
            if (ipAddrTemp.fromString(cipHttpStr))
            {
                ipAddr_public_old = ipAddr_public;
                ipAddr_public = ipAddrTemp;

                charBufp = ip2char(ipAddr_public);
                Serial.print("Public IP: ");
                Serial.println(charBufp);
                display.print(charBufp, 2, 0);
            }
            else
                Serial.println("It is a invaild address");
        }
        else
            Serial.printf("[HTTP] GET... failed, error: %s\n", cipHttp.errorToString(httpCode).c_str());
        cipHttp.end();

        /* DDNS Update */
        if (ipAddr_public == ipAddr_public_old)
            Serial.println("IP address has not changed");
        else
        {
            Serial.println("DDNS Updating...");
            Serial.print("connecting to ");
            Serial.println(ddnsServer);
            if (!ddnsClient.connect(ddnsServer, HTTPS_PORT))
                Serial.println("connection failed");
            else
            {
                if (!ddnsClient.verify(ddnsServerFingerprint, ddnsServer))
                    Serial.println("certificate doesn't match");
                else
                {
                    Serial.println("certificate matches");

                    String ddnsURL = String("/update?host=") + ddnsHost + "&domain=" + ddnsDomain + "&password=" + ddnsPassword;
                    Serial.print("requesting URL: ");
                    //Serial.println(ddnsURL);
                    Serial.println("********");        // for security

                    ddnsClient.print(String("GET ") + ddnsURL + " HTTP/1.1\r\n" +
                        "Host: " + ddnsServer + "\r\n" +
                        "User-Agent: */* \r\n" +
                        "Connection: close\r\n\r\n");

                    Serial.println("request sent");
                    while (ddnsClient.connected())
                    {
                        String line = ddnsClient.readStringUntil('\n');
                        if (line == "\r")
                        {
                            Serial.println("headers received");
                            break;
                        }
                    }
                    String ddnsReStr = ddnsClient.readStringUntil('\n');
                    Serial.println("Reply was:");
                    Serial.println("==========");
                    Serial.println(ddnsReStr);
                    Serial.println("==========");
                    Serial.println("closing connection");

                    if (ddnsReStr.indexOf("<ErrCount>0</ErrCount>") > 0)
                    {
                        Serial.println("DDNS Update Successfully");
                        display.print("DDNS Updated !", 3, 0);
                    }
                    else
                    {
                        Serial.println("DDNS Update Failed");
                        display.print("DDNS Error ...", 3, 0);
                    }
                }
            }
        }
    }

    if ((!millisHFM && millis() > funcTCnext_timeDisp) || (millisHFM && (millis() + millisHF_HalfThreshold) > (funcTCnext_timeDisp - millisHF_HalfThreshold)))
    {
        funcTCnext_timeDisp = millis() + funcTCcycle_timeDisp;
        Serial.println("Disp updating ...");
        charBufp = time2char();
        display.print(charBufp, 0, 0);
        Serial.printf("Time = %s:%02d\n", charBufp, second());

        switch (second() % 3)
        {
            case 0: display.print("/", 3, 15);
                break;
            case 1: display.print("-", 3, 15);
                break;
            case 2: display.print("\\", 3, 15);
                break;
            default:
                break;
        }
    }
    
    if ((!millisHFM && millis()>funcTCnext_millisHFM) || (millisHFM && (millis()+ millisHF_HalfThreshold)>(funcTCnext_millisHFM-millisHF_HalfThreshold)))
    {
        funcTCnext_millisHFM = millis() + funcTCcycle_millisHFM;
        if (!millisHFM && (millis() >= millisHF_UpThreshold))
            millisHFM = true;
        else if (millisHFM && (millis() >= millisHF_DownThreshold))
            millisHFM = false;
    }
    

    // wait ten seconds before asking for the time again
    //delay(10000);
}


/* Display & Serial functions */
char* ip2char(IPAddress ip)
{
    static char buf[16];
    sprintf(buf, "%3d.%3d.%3d.%3d", ip[0], ip[1], ip[2], ip[3]);
    return buf;
}

char* time2char(void)
{
    static char buf[16];
    sprintf(buf, "%4d-%02d-%02d %02d:%02d", year(), month(), day(), hour(), minute());
    return buf;
}

/* NTP functions */
// send an NTP request to the time server at the given address
unsigned long ntpSendPacket(IPAddress& address)
{
    Serial.println("sending NTP packet...");
    // set all bytes in the buffer to 0
    memset(ntpPacketBuffer, 0, NTP_PACKET_SIZE);
    // Initialize values needed to form NTP request
    // (see URL above for details on the packets)
    ntpPacketBuffer[0] = 0b11100011;    // LI, Version, Mode
    ntpPacketBuffer[1] = 0;                // Stratum, or type of clock
    ntpPacketBuffer[2] = 6;                // Polling Interval
    ntpPacketBuffer[3] = 0xEC;            // Peer Clock Precision
    // 8 bytes of zero for Root Delay & Root Dispersion
    ntpPacketBuffer[12] = 49;
    ntpPacketBuffer[13] = 0x4E;
    ntpPacketBuffer[14] = 49;
    ntpPacketBuffer[15] = 52;

    // all NTP fields have been given values, now
    // you can send a packet requesting a timestamp:
    ntpUDP.beginPacket(address, 123);        //NTP requests are to port 123
    ntpUDP.write(ntpPacketBuffer, NTP_PACKET_SIZE);
    ntpUDP.endPacket();
}

